# frozen_string_literal: trues

module Hermes
  # HousingIntention model
  class HousingApplication
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::Status
    include Mongoid::Attributes::Dynamic

    field :applicated_at, type: DateTime
    field :effort_rate, type: Float
    field :epci, type: Integer
    field :is_active, type: Boolean, default: true
    field :max_date, type: DateTime
    field :visit_conducted_date, type: DateTime
    field :last_xml_validation_report, type: String

    belongs_to :housing_offer, index: true, class_name: 'Hermes::HousingOffer'
    belongs_to :housing_request, index: true, class_name: 'Hermes::HousingRequest'
    # belongs_to :score, index: true, class_name: 'Hermes::Score'

    validates_presence_of :housing_request_id, :housing_offer

    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id

    index(applicated_at: 1)
    index(effort_rate: 1)
    index(epci: 1)
    index(is_active: 1)
    index(max_date: 1)

    before_save :before_save_actions
    after_save :after_save_actions
    after_create :after_create_actions
    after_update :after_actions
    after_destroy :after_destroy_actions

    scope :selected_by_offer, ->(housing_offer_id) {
      where(housing_offer_id: housing_offer_id, status: "TRANSMIS BAILLEUR")
    }

    HousingApplication.add_status(:single)

    def applicant
      housing_request&.main_applicant
    end

    def applicant_id
      applicant&.id
    end

    def applicated_nb_do(dec_or_inc_sym)
      return unless housing_offer

      case dec_or_inc_sym
      when :dec
        housing_offer.inc_applicated_nb(-1)
      when :inc
        housing_offer.inc_applicated_nb(1)
      end
    end

    def handle_applicated_nb
      if is_active
        applicated_nb_do :inc
      else
        applicated_nb_do :dec
      end
    end

    def update_status
      code = ""
      case status
      when "Inscription CAL"
        code = "INSCAL"
      when "Attribution lot avec rang"
        code = "DESCAL"
      when "Attribution lot"
        code = "DESCAL01"
      when "Attribution sous condition"
        code = "DESCAL02"
      when "Refus attribution"
        code = "DESCAL03"
      when "Refus lot demandeur"
        code = "REFLOG"
      when "Visite proposée"
        code = "VISPRO"
      when "Visite effectuée"
        code = "VISEFF"
      when "Radiation attribution"
        code = "RADATT"
      end
      Hermes::BoClient::UpdateStatusTask.new({
                                                 housing_application_id: id,
                                                 housing_request_id: housing_request_id,
                                                 housing_offer_id: housing_offer_id,
                                                 code_evenement: code
                                             }).execute
    end

    def update_status_update
      self.status_updated_at = DateTime.current
    end

    def before_save_actions
      update_status_update if status_changed?
    end

    def after_save_actions
      housing_request.save
      update_status if status_changed?
    end

    def after_actions
      handle_applicated_nb if is_active_changed?
    end

    # def assign_score
    #   self.score = Hermes::Score.find_by(
    #     applicant: housing_request&.main_applicant,
    #     housing_offer: housing_offer
    #   )
    # end

    def after_create_actions
      applicated_nb_do :inc if is_active
      housing_request.set(last_application_date: created_at, updated_at: DateTime.current)
      # assign_score
      # if score
      #   score.has_applicated = true
      #   score.save
      # end

      Hermes::BoClient::ApplicationTask.new({
                                                housing_application_id: id,
                                                housing_request_id: housing_request_id,
                                                housing_offer_id: housing_offer_id,
                                            }).execute
    end

    def after_destroy_actions
      housing_request.save
      applicated_nb_do :dec
    end
  end
end
