# frozen_string_literal: true

module Hermes
  class HousingRequest
    require "securerandom"
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::AlsIdConcern
    include Hermes::Concerns::Status
    include Hermes::Concerns::ReservationHolder

    hermes_tools 'revenues'

    ENTITY_STATUS_MAP = {
        spent_on_completed_requests: [
            'Salarié logé',
            'SALARIE LOGE'
        ],
        spent_on_ongoing_requests: [
            'En cours de saisie',
            'EN COURS DE SAISIE',
            'A valider',
            'A VALIDER',
            'En cours',
            'EN COURS',
            'Dossier présenté',
            'DOSSIER PRESTENTE',
            'Dossier accepté',
            'DOSSIER ACCEPTE'
        ]
    }.freeze

    SINGLE_PARENT_SITUATION_LIST = [
        'Célibataire',
        'C',
        'Divorcé (e)',
        'D',
        'Séparé (e)',
        'S',
        'Veuf',
        'V'
    ].freeze

    CATEGORY_COUNTED_CHILDREN = [
        'enfant à charge',
        'garde alternée',
        'E'
    ].freeze

    CHILDREN_TYPES = CATEGORY_COUNTED_CHILDREN + [
        'droit de visite'
    ].freeze

    CONJOINT_TYPES = [
        'Coinjoint (e)',
        'Marié (e)',
        'M',
        'Pacsé (e)',
        'P'
    ].freeze

    FAMILY_FIELDS = %w[
      dependents
      applicant_ids
    ].freeze

    ELLIGIBILITY_FIELDS = %w[
      reference_tax_revenues
      monthly_revenues
      is_active
      seeked_districts
      seeked_types
      priority_points_indexes
      mutation
      mobility
      priority_public
      single_parent
      seeked_departments
      entity_validated
      disability_details
      housing_offer_ids
      applicant_ids
      category_type
    ].freeze

    REVENUES_FIELDS = %w[
      dependents
      main_applicant_id
      applicant_ids
    ].freeze

    ORIENTATION_BOOLS = %w[
      single_parent
      priority_public
      mutation
    ].freeze

    YOUNG_COUPLE_THRESHOLD = 55
    APPROXIMATE_YEAR_SECONDS = 31_557_600

    field :is_active, type: Boolean, default: false
    field :als_id, type: String
    field :attributed_housing_district, type: String
    field :attributed_housing_district_postal_code, type: String
    field :dalo, type: Boolean
    field :asked_mobility, type: Boolean
    field :valid_mobility, type: Boolean
    field :decider, type: String
    field :submitted_to_cal_nb, type: Integer
    field :sent_to_cal_nb, type: Integer
    field :rejected_in_cal_nb, type: Integer
    field :rejected_after_cal_nb, type: Integer
    field :accepted_in_cal_nb, type: Integer
    field :mobility, type: Boolean
    field :seeked_types, type: Array
    field :entity_help_required, type: Boolean
    field :contract_type, type: String
    field :registration_number, type: String
    field :entity_validated, type: Boolean
    field :priority_points_indexes, type: Array, default: []
    field :has_priority, type: Boolean, default: false
    field :total_priority_points, type: Integer
    field :nud, type: String
    field :nb_of_requests_made, type: Integer
    field :reference_tax_revenues, type: Float
    field :dependents, type: Array
    field :category_type, type: Integer
    field :single_parent, type: Boolean
    field :id_counter, type: String
    field :id_unique, type: String
    field :id_unique_associate, type: String
    field :internal_ref, type: String
    field :forcing_duplicate, type: Boolean
    field :electronical_renewal, type: Boolean
    field :ec_law_ressources, type: String
    field :is_long_delay_anormal, type: Boolean
    field :date_calculator_indicator, type: DateTime
    field :with_car_park, type: Boolean
    field :widen_request, type: Boolean
    field :housing_type_list, type: Array #Enum
    field :expulsion_judgement_date, type: DateTime
    field :other_reason, type: String
    field :work_contract_type, type: Array #Enum
    field :seeked_districts, type: Array, default: []
    field :seeked_district_postal_codes, type: Array, default: []
    field :seeked_departments, type: Array, default: []
    field :truly_seeked_departments, type: Array, default: []
    field :disability_details, type: Array
    field :seeked_localisation_list, type: Array
    field :profession_details, type: Hash

    field :requester_firstname, type: String
    field :requester_lastname, type: String
    field :requester_fullname, type: String
    field :requester_email, type: String
    field :requester_birth_date, type: DateTime

    field :monthly_revenues, type: Float
    field :entity_id, type: String
    field :typology_range, type: Array #Enum
    field :family_size, type: Integer
    field :unborn_children, type: Array
    field :lessor_name, type: String
    field :lessor_siren, type: String
    field :social_housing_entry_date, type: Date
    field :social_housing_name, type: String
    field :test_created_at, type: DateTime
    field :sne_number, type: String
    field :actual_housing_details, type: Hash
    field :postal_details, type: Hash
    field :owned_housing_details, type: Hash
    field :employee_id, type: String
    field :employee_start_date, type: DateTime
    field :entity_access_level, type: String
    field :employee_unit, type: String #Enum
    field :tab_state, type: String
    field :motivation1, type: String
    field :motivation2, type: String
    field :motivation3, type: String
    field :seeking_first_floor, type: Boolean
    field :seeking_elevator, type: Boolean
    field :seeked_max_rent, type: Integer
    field :seeked_epci_first_quartils, type: Array
    field :seeked_quartil_belongings, type: Array
    field :sne_creation_date, type: DateTime
    field :precisions, type: String
    field :has_multiple_attributions_under_conditions, type: Boolean
    field :seeked_housing_categories, type: Array #Enum
    field :min_surface, type: Integer, default: 0
    field :priority_public, type: Boolean
    field :mutation, type: Boolean
    field :dependents_monthly_incomes, type: Hash, default: {}
    field :dependents_monthly_charges, type: Hash, default: {}
    field :radiation_date, type: Date
    field :radiation_motive, type: String
    field :valid_seeked_types, type: Array, default: []
    field :out_of_range_types, type: Array, default: []
    field :applications_nb, type: Integer, default: 0
    field :last_application_date, type: DateTime
    field :housing_application_ids_was, type: Array
    field :is_house_owner, type: Boolean
    field :nud_ids_was, type: Array

    # APPLICANT MODIFIED
    # serve as triggers for before_save and after_save

    field :name_has_changed, type: Boolean
    field :birth_date_has_changed, type: Boolean
    field :monthly_revenues_will_change, type: Boolean
    field :category_may_change, type: Boolean
    field :should_update_eligibility, type: Boolean
    field :will_update_status_dependencies, type: Boolean
    field :tax_revenues_will_change, type: Boolean

    field :superfields, type: Hash, default: {}

    field :applicant_ids, type: Array, default: []
    field :housing_offer_ids, type: Array, default: []
    field :documents_required, type: Integer
    field :documents_uploaded, type: Integer
    field :workplace_insee_code, type: String
    field :nud_creations, type: Hash, default: {} # Hash of form : {department1 => oldest_creation_date, department2 => oldest_creation_date, ...}

    field :last_xml_validation_report, type: String


    validates_presence_of :main_applicant_id

    attr_accessor *%i[
      only_public
      effort_rate
      main_user_id
    ]

    belongs_to(
        :main_applicant,
        class_name: 'Hermes::Applicant',
        index: true
    )
    belongs_to :housing_request_folder, class_name: "Hermes::HousingRequestFolder", index: true

    # faire un alias en main_applicant && main_applicant_id
    # has_many :applicants, class_name: 'Hermes::Applicant' # faire une alias en coapplicants && coapplicant_ids
    # pour les coapplicants => il faut un moyen de stocker leur relation avec le demandeur principale
    # de type : 'connjoint', 'paxe', 'colocataire', etc ... => trouver un moyen de gerer cette relation
    # modele intermediaire / de jointure peut-être ?
    has_many :housing_applications, class_name: 'Hermes::HousingApplication'
    has_many :nuds, class_name: 'Hermes::Nud'

    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id

    common_indexes *%i[
      is_active
      als_id
      dalo
      submitted_to_cal_nb
      sent_to_cal_nb
      rejected_in_cal_nb
      rejected_after_cal_nb
      accepted_in_cal_nb
      seeked_types
      registration_number
      entity_validated
      priority_points_indexes
      has_priority
      total_priority_points
      nud
      reference_tax_revenues
      seeked_districts
      category_type
      single_parent
      id_counter
      with_car_park
      widen_request
      seeked_district_postal_codes
      seeked_departments
      seeked_localisation_list
      requester_firstname
      requester_lastname
      requester_fullname
      requester_birth_date
      monthly_revenues
      entity_id
      typology_range
      family_size
      employee_id
      entity_access_level
      seeking_first_floor
      seeking_elevator
      seeked_max_rent
      sne_creation_date
      seeked_housing_categories
      min_surface
      priority_public
      mutation
      housing_offer_ids
      housing_application_ids
      nud_ids
      nud_creations
    ]

    scope :actives, -> {where(is_active: true)}

    before_create :before_create_actions
    before_save :before_save_actions
    after_create :after_create_actions
    before_update :before_actions #, https://dev.hermes.openstack.local/admin/families/5cffa87f4ac06e526bd93295/mapping
    after_update :after_actions
    after_save :after_save_actions

    HousingRequest.add_status(:single, 'entity_status', 'EN COURS DE SAISIE')
    HousingRequest.add_status(:single, 'request_status')

    def entity
      main_applicant&.entity
    end

    def housing_offers
      Hermes::HousingOffer.where(:id.in => housing_offer_ids)
    end

    def coapplicants_raw
      Hermes::Applicant.where(:id.in => (applicant_ids || []))
    end

    def coapplicants
      return [] unless applicant_ids
      applicant_ids.map do |aid|
        if applicant_link = Hermes::ApplicantLink.unscoped.find_by("$or"=>[{applicant_ids: aid.to_s}, {applicant_ids: aid}], housing_request_id: self.id)
          {rel_type: applicant_link.type, id: aid.to_s}
        end
      end
    end

    def coapplicants=(arr)
      return unless arr.is_a?(Array)

      self.superfields[:last_coapplicants] = arr

      present_ids = arr.map {|elt| elt['id'] || elt[:id]}
      missing_ids = applicant_ids.to_array - present_ids.to_array

      missing_ids.each do |missing_id|
       # ap missing_id
        link = Hermes::ApplicantLink.find_by("$or"=>[{applicant_ids: missing_id.to_s}, {applicant_ids: missing_id}], housing_request_id: id)
        next unless link

        link.applicant_ids.delete(missing_id)
        link.applicant_ids.delete(missing_id.to_s)
        link.save
      end

      arr.each do |h|
        next unless h.is_a?(Hash) && h.present?
        next unless Hermes::Applicant.find_by(id: (h[:id] || h['id']))

        rel_type = h[:rel_type] || h['rel_type']
        applicant_id = h[:id] || h['id']
        other_link = Hermes::ApplicantLink.find_by(applicant_ids: applicant_id, housing_request_id: id)
        link = Hermes::ApplicantLink.find_or_create_by(type: rel_type, housing_request_id: id)
        if link && other_link && link.id != other_link.id
          other_link.applicant_ids.delete(applicant_id)
          other_link.applicant_ids.delete(applicant_id.to_s)
          other_link.save
        end
        link.applicant_ids << applicant_id
        link.save
      end
      self.applicant_ids = Hermes::ApplicantLink.where(housing_request_id: id)
                               .pluck(:applicant_ids)
                               .flatten
                               .uniq
                               .compact
    end

    def update_valid_seeked_types
      formated_range = (typology_range || [])
      valid_types = (seeked_types || []) & formated_range
      self.valid_seeked_types = valid_types.blank? ? nil : valid_types
      self.out_of_range_types = valid_types - formated_range | formated_range - valid_types
    end

    def inherit_acl
      %w[readers writers].each {|acl| send("#{acl}=", main_applicant&.send(acl))}
    end

    def inherit_applicant_datas
      # TODO: set to false after process
      return unless main_applicant

      self.name_has_changed = false
      self.requester_firstname = main_applicant.first_name
      self.requester_lastname = main_applicant.last_name
      self.requester_fullname = main_applicant.full_name
      self.requester_email = main_applicant.email
      inherit_acl if is_active
    end

    def inherit_applicant_birth_date
      return unless main_applicant

      self.birth_date_has_changed = false
      self.requester_birth_date = main_applicant.birth_date
    end

    def handle_active_request
      return unless main_applicant

      if is_active
        other_active_requests = main_applicant.housing_requests.actives
                                    .where(id: {:$ne => id})
        other_active_requests.set(is_active: false)
      elsif main_applicant.housing_requests.actives.count.zero?
        update_applicant_eligbility
      end
    end

    def main_applicant_revenues
      main_applicant&.monthly_revenues || 0
    end

    def coapplicants_monthly_revenues
      (coapplicants_raw || []).inject(0) do |sum, coapplicant|
        sum + coapplicant.monthly_revenues
      end
    end

    def dependents_monthly_revenues
      (dependents || []).inject(0) do |sum, dependent|
        dependent_incomes = (dependent['monthly_incomes']&.values_fsum || 0)
        dependent_charges = (dependent['monthly_charges']&.values_fsum || 0)
        sum + dependent_incomes - dependent_charges
      end
    end

    def update_total_priority_points
      self.total_priority_points = (priority_points_indexes || []).sum
    end

    def update_monthly_revenues
      self.monthly_revenues_will_change = false
      self.monthly_revenues = main_applicant_revenues +
          coapplicants_monthly_revenues +
          dependents_monthly_revenues
    end
   def couple_type
      if single_parent
        'single_parent'
      elsif coapplicants.any? do |coapplicant|
        next unless coapplicant

        CONJOINT_TYPES.include?(coapplicant['rel_type'])
      end
        'couple'
      else
        'single'
      end
    end

    def conjoint
      conjoint_id = coapplicants.find do |coapplicant|
        next unless coapplicant

        CONJOINT_TYPES.include?(coapplicant['rel_type'])
      end&.fetch('id', nil)
      return unless conjoint_id

      Hermes::Applicant.find_by(id: coinjoint_id)
    end

    def young_couple?
      if conjoint
        births = [main_applicant, conjoint].map(&:birth_date)
        current_time = Time.current
        births_sum = births.inject(0) do |sum, birth|
          next unless birth.is_a?(String) && birth.present?

          sum + current_time - Time.parse(birth)
        end
        births_sum <= YOUNG_COUPLE_THRESHOLD * APPROXIMATE_YEAR_SECONDS
      else
        false
      end
    end

    def handicap?
      (disability_details || []).any? {|disability| disability.is_a?(Hash) && disability['mobility_inclusion_card'] == true}
    end

    def coapplicant_nb
      applicant_ids.present? ? coapplicants_raw.count : 0
    end

    def limited_dependent_nb
      (dependents || []).select do |dependent|
        next unless dependent.present?

        rel_type = dependent['rel_type'] || dependent[:rel_type]
        CATEGORY_COUNTED_CHILDREN.include?(rel_type)
      end.length
    end

    def children_dependent_nb
      (dependents || []).select do |dependent|
        next unless dependent.present?

        rel_type = dependent['rel_type'] || dependent[:rel_type]
        CHILDREN_TYPES.include?(rel_type)
      end.length
    end

    def less_limited_dependent_nb
      (dependents || []).select {|dependent| dependent.present?}.length
    end

    def limited_family_nb
      coapplicant_nb + limited_dependent_nb
    end

    def calculate_category
      first_cond = limited_family_nb >= 3 ||
          (single_parent && limited_dependent_nb > 1) ||
          (handicap? && limited_family_nb > 1)
      second_cond = limited_family_nb == 2 ||
          (
          single_parent &&
              coapplicant_nb.zero? &&
              limited_dependent_nb == 1
          ) ||
          (young_couple? && limited_family_nb.zero?) ||
          (handicap? && limited_family_nb == 1) ||
          (coapplicant_nb == 1 && children_dependent_nb == 1)
      third_cond = (coapplicant_nb == 1 && limited_dependent_nb.zero?) ||
          young_couple? ||
          handicap? ||
          (coapplicant_nb.zero? && children_dependent_nb == 1)

      if first_cond
        if handicap?
          limited_family_nb + 2
        elsif single_parent && (category = limited_dependent_nb + 2) > limited_family_nb
          category
        else
          limited_family_nb + 1
        end
      elsif second_cond
        3
      elsif third_cond
        2
      else
        1
      end
    end

    def update_category
      self.category_type = main_applicant.nil? ? nil : calculate_category
    end

    def update_family_size
      self.family_size = coapplicant_nb + less_limited_dependent_nb + 1
    end

    def calculate_min_surface
      case coapplicant_nb
      when 0
        9 +
            (
            (less_limited_dependent_nb.positive? ? 1 : 0) *
                (
                7 +
                    ((less_limited_dependent_nb - 1) * 9) -
                    (
                    (less_limited_dependent_nb > 7 ? 1 : 0) *
                        (less_limited_dependent_nb - 7)
                    )
                )
            )
      else
        16 + (9 * less_limited_dependent_nb)
      end
    end

    def update_applications_nb
      self.applications_nb = housing_applications.count
    end

    def update_housing_application_ids_was
      set(housing_application_ids_was: housing_application_ids)
    end

    def housing_application_ids_changed?
      formated_ids = (housing_application_ids || [])
      formated_ids_was = (housing_application_ids_was || [])
      (
      (housing_application_ids && !housing_application_ids_was) &&
          (housing_application_ids_was && !housing_application_ids)
      ) ||
          (formated_ids.length != formated_ids_was.length) ||
          !(
          formated_ids.all? {|elt| formated_ids_was.include?(elt)} &&
              formated_ids_was.all? {|elt| formated_ids.include?(elt)}
          )
    end

    def update_update_nud_ids_was
      set(housing_application_ids_was: housing_application_ids)
    end

    def nud_ids_changed?
      formated_ids = (nud_ids || [])
      formated_ids_was = (nud_ids_was || [])
      (
      (nud_ids && !nud_ids_was) &&
          (nud_ids_was && !nud_ids)
      ) ||
          (formated_ids.length != formated_ids_was.length) ||
          !(
          formated_ids.all? {|elt| formated_ids_was.include?(elt)} &&
              formated_ids_was.all? {|elt| formated_ids.include?(elt)}
          )
    end

    def update_min_surface
      self.min_surface = calculate_min_surface
    end

    def update_tax_revenues
      self.tax_revenues_will_change = false
      self.reference_tax_revenues = main_applicant&.reload.amount_year_minus_two
    end

    def calculate_typology_range
      dependent_nb = less_limited_dependent_nb
      min, max = case coapplicant_nb
                 when 0
                   [
                       if dependent_nb < 2
                         dependent_nb + 1
                       else
                         dependent_nb < 8 ? dependent_nb : 6
                       end,
                       2 + dependent_nb
                   ]
                 else
                   [
                       (dependent_nb < 5 ? dependent_nb + 1 : 6),
                       3 + dependent_nb
                   ]
                 end
      self.typology_range = (min..max).map {|i| "T#{i}"}
    end

    def update_typology_range
      self.typology_range = calculate_typology_range
    end

    def bool_filter
      ORIENTATION_BOOLS.each_with_object({}) do |bool_name, h|
        bool = send(bool_name)
        h[bool_name] = {:$ne => true} if bool.class.name != 'TrueClass'
      end
    end

    def reserved_filter
      if entity && entity_validated
        {
            :$or => [
                {reserved: true, entity_id: entity_id.to_s},
                {reserved: {:$ne => true}}
            ]
        }
      else
        {reserved: {:$ne => true}}
      end
    end

    def category_filter
      filter = if reference_tax_revenues
                 {
                     "ceil_category_#{category_type}".to_sym => {:$gte => reference_tax_revenues},
                     "floor_category_#{category_type}".to_sym => {:$lte => reference_tax_revenues}
                 }
               else
                 self.only_public = true
                 {}
               end
      self.category_may_change = false
      filter
    end

    def size_filter
      filter = {max_family_size: {:$gte => family_size}}
      field = if single_parent
                :single_min_family_size
              else
                :couple_min_family_size
              end
      filter.merge(field => {:$lte => family_size})
    end

    def typology_filter
      filter = []
      valid_types = (valid_seeked_types || []).map(&:capitalize)
      valid_types = typology_range.to_a if valid_types.blank?
      valid_types.each do |type|
        filter[0] ||= {:$or => []}
        filter[0][:$or] << {typology: type}
      end
      filter
    end

    def min_revenues_filter
      if monthly_revenues
        filter = {min_revenues: {:$lte => (monthly_revenues&.to_f || 0.0)}}
      else
        self.only_public = true
      end
      filter || {}
    end

    def geo_filter
      filter = {}
      filter[:department] = {:$in => truly_seeked_departments} if truly_seeked_departments.present?
      filter
    end

    def eligibility_filter
      if only_public || !is_active
        self.only_public = false
        return {is_public: true}
      end

      filter = bool_filter.deep_merge!(reserved_filter)
                   .deep_merge!(size_filter)
                   .deep_merge!(min_revenues_filter)
                   .deep_merge!(geo_filter)
                   .deep_merge!(category_filter)
      [typology_filter].each do |or_filter|
        next unless or_filter.present?

        filter[:$and] ||= []
        filter[:$and].concat(or_filter)
      end
      filter
    end

    def score_all
      return unless main_applicant
      offer_ids = Hermes::HousingOffer.where(
          **main_applicant.eligibility_query.json_to_h.merge(offer_status: { :$in => HousingOffer.published_statuses })
      ).pluck(:id).map(&:to_s)
      return unless offer_ids.present?

      UpdateScoreJob.perform_later('housing_request', offer_ids, main_applicant.id.to_s)
    end

    def update_applicant_eligbility
      return unless main_applicant

      query = eligibility_filter.to_json
      if query != main_applicant.eligibility_query
        main_applicant.set(eligibility_query: query, updated_at: DateTime.current)
        score_all
      end
      set(should_update_eligibility: false)
    end

    def priority_points?
      priority_points_indexes&.length&.positive?
    end

    def update_has_priority
      self.has_priority = priority_points?
    end

    def entity_status_to_map
      ENTITY_STATUS_MAP.find {|_, arr| arr.include?(entity_status)}&.first ||
          'spent_on_other'
    end

    def update_entity_stock
      return unless entity.present? && entity_validated

      points_now = (priority_points_indexes || []).sum
      points_before = (priority_points_indexes_was || []).sum
      points_difference = points_now - points_before

      entity.send(
          "#{entity_status_to_map}=",
          (entity.send(entity_status_to_map) || 0) + points_difference
      )
      entity.save
    end

    def single_parent_situation?
      SINGLE_PARENT_SITUATION_LIST.include?(main_applicant&.family_situation)
    end

    def category_any_counted_children?
      return false unless dependents.present?

      dependents.any? do |dependent|
        CATEGORY_COUNTED_CHILDREN.include?(dependent[:rel_type])
      end
    end

    # handle nud_creations field - Hermes::Nud dependent
    def handle_nud_creations
      # for each unique department present for associated nuds
      # set nud_creations Hash with the oldest date for the defined department
      present_departments = nuds.pluck(:department).uniq.compact
      present_departments.each do |department|
        self.nud_creations ||= {}
        self.nud_creations[department] = nuds.where(department: department)
                                             .min(:creation_date)
      end
    end

    def update_single_parent
      self.single_parent = if single_parent_situation?
                             category_any_counted_children?
                           else
                             false
                           end
      self.category_may_change = single_parent_changed?
    end

    def update_seeked_location
      seeked_localisation_list.each do |seeked_localisation|
        check_seeked_localisation(seeked_localisation)
        seeked_localisation= seeked_localisation.stringify_keys if seeked_localisation.is_a? Hash

        seeked_localisation['codesPostaux']&.each do |postal_code|
          if postal_code.present?
            seeked_district_postal_codes.push(postal_code)
            truly_seeked_departments.push(postal_code[0..1])
          end
        end

        if seeked_localisation['codeDepartement']
          seeked_departments.push(seeked_localisation['codeDepartement'])
          truly_seeked_departments.push(seeked_localisation['codeDepartement'])
        end

        seeked_districts.push(seeked_localisation['nom']) if seeked_localisation['nom']
      end

      %i[seeked_district_postal_codes seeked_departments seeked_districts truly_seeked_departments].each do |arr|
        self[arr] = self[arr].uniq.sort!
      end
    end

    def update_housing_offers
      inter = self.housing_offer_ids = housing_applications.map do |application|
        application.housing_offer&.id.to_s
      end
      self.housing_offer_ids = inter.uniq.compact
    end

    def any_changed_for?(constant_list)
      constant_list.any? {|field| send("#{field}_changed?")}
    end

    def should_update_family_size?
      any_changed_for?(FAMILY_FIELDS)
    end

    def should_update_eligibility?
      should_update_eligibility ||
          any_changed_for?(ELLIGIBILITY_FIELDS)
    end

    def monthly_revenues_will_change?
      monthly_revenues_will_change ||
          any_changed_for?(REVENUES_FIELDS)
    end

    def housing_differences?
      return false if housing_offer_ids.blank? && housing_application_ids.blank?

      formated_offer_ids = housing_offer_ids.to_array
      formated_application_ids = housing_application_ids.to_array
      echelon = formated_offer_ids.length
      return true if echelon != formated_application_ids.length

      (formated_offer_ids & formated_application_ids).length != echelon
    end

    def generate_als_id
      last_housing_request = Hermes::HousingRequest.where(als_id: { '$ne' => nil }).order_by(als_id: :desc).first
      start = '%010d' % 1

      if last_housing_request
        self.als_id = '%010d' % (last_housing_request.als_id.to_i + 1)
      else
        self.als_id = start
      end
    end

    # def orientation_changes
    #   changes = ORIENTATION_BOOLS.each_with_object({}) do |orientation_name, h|
    #     if send("#{orientation_name}_changed?")
    #       h[orientation_name.to_sym] = if send(orientation_name)
    #                                      -1
    #                                    else
    #                                      1
    #                                    end
    #     end
    #   end
    #   if entity_id_changed?
    #     changes[:entities] = {
    #       entity_id_was.to_sym => -1,
    #       entity_id.to_sym => 1
    #     }
    #   end
    #   changes
    # end

    # def orientation_changed?
    #   orientation_changes.present?
    # end

    # def handle_orientation_movements
    #   return unless eligibility_query && eligibility_query != "{is_public:true}"

    #   HandleRequestOrientationMovementJob.perform_later(id.to_s, orientation_changes)
    # end
    def  check_seeked_localisation(seeked_localisation)
      if( !(seeked_localisation["nom"].is_a? String) ||!(seeked_localisation["code"].is_a? String) || !(seeked_localisation["codesPostaux"].is_a? Array)|| !(seeked_localisation["codeDepartement"].is_a? String))
        self.errors.add(:base, "non-conform seeked_localisation_list, it should looks like [{'nom':'paris', 'code':'75113', 'codesPostaux':['75013',... },{'codeDepartement':'75'},{...},...]")
      end
    end

    def touch_scores
      UpdateScoreJob
    end

    def before_create_actions
      generate_als_id
      inherit_applicant_datas
      inherit_applicant_birth_date
      handle_nud_creations if nud_ids_changed?
      update_applications_nb
      update_tax_revenues
      update_min_surface
      update_family_size
      update_typology_range
      update_valid_seeked_types
      if entity_validated && priority_points_indexes_changed?
        update_has_priority
        update_total_priority_points
      end
      update_single_parent
      update_monthly_revenues
      update_category
      update_housing_offers
    end

    def before_save_actions
      update_seeked_location if seeked_localisation_list_changed?
    end

    def before_actions
      inherit_applicant_datas if name_has_changed
      inherit_applicant_birth_date if birth_date_has_changed
      handle_nud_creations if nud_ids_changed?
      update_applications_nb if housing_application_ids_changed?
      update_tax_revenues if tax_revenues_will_change
      if should_update_family_size?
        update_min_surface
        update_family_size
        update_typology_range
      end
      update_valid_seeked_types if seeked_types_changed? || typology_range_changed?
      if entity_validated && priority_points_indexes_changed?
        update_has_priority
        update_total_priority_points
      end
      update_single_parent # TODO: if dependents_changed? || category_may_change
      update_monthly_revenues if monthly_revenues_will_change?
      update_category # TODO: if category_may_change
      update_housing_offers if housing_differences?
    end

    def after_create_actions
      update_applicant_eligbility
    end

    def update_main_applicant_main_user
      return unless main_applicant

      main_applicant.main_user_id = main_user_id
      main_applicant.save

      self.main_user_id = nil
    end

    def update_housing_request_folder
      if (main_applicant_id_changed? ||
          applicant_ids_changed? ||
          motivation1_changed? ||
          motivation2_changed? ||
          motivation3_changed? ||
        unborn_children_changed? ||
        dependents_changed? ||
        housing_type_list_changed? )
        housing_request_folder.compute_required_documents
        housing_request_folder.save
      end
    end

    #update only
    def after_actions
      update_housing_application_ids_was if housing_application_ids_changed?
      update_nud_ids_was if housing_application_ids_changed?
      priority_points_indexes_changed?
      handle_active_request if is_active_changed?
      update_applicant_eligbility if should_update_eligibility? || is_active_changed?

    end

    # WARNING : can produce loops if miss-used
    def update_applicant_for_nuds
      nuds.each do |nud|
        nud.applicant_id = main_applicant_id
        nud.save
      end
    end

    #every save
    def after_save_actions
      check_status_changed
      can_update_entity_stock = entity &&
          entity_validated &&
          priority_points_indexes.present?
      handle_active_request if is_active
      update_entity_stock if can_update_entity_stock

      update_applicant_for_nuds if main_applicant_id_changed?
      update_main_applicant_main_user if main_user_id

      if entity_validated_changed?
        Hermes::BoClient::AddRequestTask.new(housing_request_id: id).execute
      end

      set(housing_request_folder_id: Hermes::HousingRequestFolder.create(housing_request_id: id).id, updated_at: DateTime.current) unless housing_request_folder_id
      update_housing_request_folder
      if request_status === "ELIGIBLE" || request_status === "COMPLETE"
        if !nud_ids || nud_ids.length == 0
          Hermes::BoClient::AddRequestTask.new(housing_request_id: id, nud_id: false).execute
        else
          nud_ids.each do |nud_id|
            Hermes::BoClient::AddRequestTask.new(housing_request_id: id, nud_id: nud_id).execute
          end
        end
      end
      return true
    end
  end
end